/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "tawCalculatePolarConfig.h"

#include <runtimeInfo/runtimeInfo.h>
#include <moduleBasics/module.h>

#include <algorithm>
#include <exception>
#include <iomanip>
#include <string>
#include <vector>
#include <utility>  // make_pair
#include <iostream>
#include <sstream>

#ifdef _WIN32
    #include <fileapi.h>
    #include <windows.h>
#endif  // _WIN32

using std::endl;

tawCalculatePolarConfig::tawCalculatePolarConfig(const node& config)
    :
    // execution mode
    execution_mode(EndnodeReadOnly<std::string>("program_settings/execution_mode").read(config)),
    // lifting line settings
    stepWidthCL(EndnodeReadOnly<double>("LiftingLineForTAW/stepWidthCL").read(config)),
    // other settings
    CLModeAirfoilSelection(EndnodeReadOnly<std::string>("program_settings/CLModeAirfoilSelection").read(config)),
    setFlightConditionsMode(EndnodeReadOnly<bool>("program_settings/FlightConditions/AdaptMachNumbersToCruiseReq").read(config).value()),
    pathToLiftingLine(EndnodeReadOnly<std::string>("program_settings/LiftingLine/FolderPath").read(config)),
    // lifting line for TAW pitching moment corrections
    CM_corr_fuselage_mode(EndnodeReadOnly<std::string>("program_settings/LiftingLineForTAW/PitchingMoment/CM_corr_fuselage").read(config)),
    CM_corr_nacelle_mode(EndnodeReadOnly<std::string>("program_settings/LiftingLineForTAW/PitchingMoment/CM_corr_nacelle").read(config)),
    delta_CM0(EndnodeReadOnly<double>("LiftingLineForTAW/PitchingMoment/delta_CM0").read(config)),
    delta_dCMdCL(EndnodeReadOnly<double>("LiftingLineForTAW/PitchingMoment/delta_dCMdCL").read(config)),
    // lifiting line for TAW induced drag corrections
    indDragCtCorrForCalibration(EndnodeReadOnly<double>("LiftingLineForTAW/InducedDragCorrection/indDragCtCorrForCalibration").read(config)),
    factorIndDragCleanPolar(EndnodeReadOnly<double>("LiftingLineForTAW/InducedDragCorrection/factorIndDragCleanPolar").read(config)),
    // visc drag raymer transition parameters
    manualTransition(EndnodeReadOnly<bool>("program_settings/ViscDragRaymer/ManualTransition").read(config)),
    manualTransitionWing(EndnodeReadOnly<double>("program_settings/ViscDragRaymer/TransitionLocationWing").read(config)),
    manualTransitionStabilizer(EndnodeReadOnly<double>("program_settings/ViscDragRaymer/TransitionLocationStabilizer").read(config)),
    manualTransitionFin(EndnodeReadOnly<double>("program_settings/ViscDragRaymer/TransitionLocationFin").read(config)),
    // visc drag raymer calibration parameters
    doViscDragCalibration(EndnodeReadOnly<bool>("program_settings/ViscDragRaymer/UseCalibration").read(config)),
    CDvisc_CDsum(EndnodeReadOnly<double>("program_settings/ViscDragRaymer/CalibrationHighMa/CDSum").read(config)),
    CDvisc_CLfact(EndnodeReadOnly<double>("program_settings/ViscDragRaymer/CalibrationHighMa/CLFact").read(config)),
    CDvisc_CLexp(EndnodeReadOnly<double>("program_settings/ViscDragRaymer/CalibrationHighMa/CLExp").read(config)),
    CDvisc_CDsum_lowMa(EndnodeReadOnly<double>("program_settings/ViscDragRaymer/CalibrationLowMa/CDSum").read(config)),
    CDvisc_CLfact_lowMa(EndnodeReadOnly<double>("program_settings/ViscDragRaymer/CalibrationLowMa/CLFact").read(config)),
    CDvisc_CLexp_lowMa(EndnodeReadOnly<double>("program_settings/ViscDragRaymer/CalibrationLowMa/CLExp").read(config)),
    doViscousDragRaymerDragCountCorrection(EndnodeReadOnly<bool>("program_settings/ViscDragRaymer/DoDragCountCorrection").read(config)),
    deltaViscousDragCountWing(EndnodeReadOnly<double>("program_settings/ViscDragRaymer/DeltaViscousDragWing").read(config)),
    deltaViscousDragCountFuselage(EndnodeReadOnly<double>("program_settings/ViscDragRaymer/DeltaViscousDragFuselage").read(config)),
    deltaViscousDragCountHTP(EndnodeReadOnly<double>("program_settings/ViscDragRaymer/DeltaViscousDragHTP").read(config)),
    deltaViscousDragCountVTP(EndnodeReadOnly<double>("program_settings/ViscDragRaymer/DeltaViscousDragVTP").read(config)),
    deltaViscousDragCountNacelles(EndnodeReadOnly<double>("program_settings/ViscDragRaymer/DeltaViscousDragNacelles").read(config)),
    deltaViscousDragCountMisc(EndnodeReadOnly<double>("program_settings/ViscDragRaymer/DeltaViscousDragMisc").read(config)),
    // wave drag mason
    numberWingStrips(EndnodeReadOnly<int>("program_settings/WaveDragMason/NumberWingStrips").read(config)),
    useCostomReferenceAngleForSweep(EndnodeReadOnly<bool>("program_settings/WaveDragMason/UseCostomReferenceAngleForSweep").read(config)),
    Ka(EndnodeReadOnly<double>("program_settings/WaveDragMason/Ka").read(config)),
    maximumSegmentSweep(EndnodeReadOnly<double>("program_settings/WaveDragMason/MaximumSegmentSweep").read(config)),
    customSweepPosition(EndnodeReadOnly<double>("program_settings/WaveDragMason/CustomSweepAngle").read(config)),
    doWaveDragMasonCalibration(EndnodeReadOnly<bool>("program_settings/WaveDragMason/UseCalibration").read(config)),
    CDwave_CLFact(EndnodeReadOnly<double>("program_settings/WaveDragMason/CLFact").read(config)),
    CDwave_CLExp(EndnodeReadOnly<double>("program_settings/WaveDragMason/CLExp").read(config)),
    doWaveDragMasonDragCountCorrection(EndnodeReadOnly<bool>("program_settings/WaveDragMason/DoDragCountCorrection").read(config)),
    deltaWaveDragCountWing(EndnodeReadOnly<double>("program_settings/WaveDragMason/DeltaWaveDragWing").read(config)),
    deltaWaveDragCountStabilizer(EndnodeReadOnly<double>("program_settings/WaveDragMason/DeltaWaveDragStabilizer").read(config)),
    // trim settings
    doTrim(EndnodeReadOnly<bool>("program_settings/Trim/DoTrim").read(config)),
    trimAdditionalCGPositions(EndnodeReadOnly<bool>("program_settings/Trim/TrimWithAdditionalCoGPositions").read(config)),
    trimHighLiftPolars(EndnodeReadOnly<bool>("program_settings/Trim/TrimHighLift").read(config)),
    stabAngleGridMode(EndnodeReadOnly<std::string>("program_settings/Trim/StabAngleGrid").read(config)),
    customStabAngleGrid(EndnodeReadOnly<std::string>("program_settings/Trim/customStabAngleGrid").read(config)),
    design_trim_MTOM_fraction(EndnodeReadOnly<double>("program_settings/Trim/design_trim_MTOM_fraction").read(config).value()),
    deltaTotalDragHL(EndnodeReadOnly<double>("program_settings/SemiEmpiricalHighLiftAdaptions/deltaTotalDrag").read(config)),
    factorDragHL(EndnodeReadOnly<double>("program_settings/SemiEmpiricalHighLiftAdaptions/factorDrag").read(config)),
    // high lift polars settings
    do_individual_hl_calibration(EndnodeReadOnly<bool>("program_settings/SemiEmpiricalHighLiftAdaptions/do_individual_hl_calibration").read(config).value()),
    apply_manual_CLmax_hl_values(EndnodeReadOnly<bool>("program_settings/SemiEmpiricalHighLiftAdaptions/apply_manual_CLmax_hl_values").read(config).value()),
    autoFC_allowGridChange(EndnodeReadOnly<bool>("program_settings/FlightConditions/allow_grid_change").read(config).value()),
    autoFC_extrapolationMargin(EndnodeReadOnly<double>("program_settings/FlightConditions/extrapolation_margin").read(config).value()) {
    // Constructor
    liftingLineFolderPath = pathToLiftingLine.value();
    myRuntimeInfo->out << "config: " << liftingLineFolderPath << endl;
    myRuntimeInfo->info << "set flight conditions from config xml" << endl;
    size_t numberFlightConditions(config.getVector(("program_settings/FlightConditions/FlightCondition"), 2).size());
    for (size_t i(0); i < numberFlightConditions; i++) {
        readFlightConditions(config, i);
    }
    // Sort flight conditions asc by Mach numbers
    std::sort(flightConditions.begin(), flightConditions.end(),
                [](const flightCondition& firstFC, const flightCondition& secondFC) {
                    return firstFC.Mach <= secondFC.Mach;
                });
    this->initializeTrimSettings(config);
    this->read_hl_calibration_factors(config);
}

void tawCalculatePolarConfig::readFlightConditions(const node& config, size_t ID) {
    EndnodeReadOnly flightConditionExtrapolationMargin(EndnodeReadOnly<double>("program_settings/FlightConditions/FlightCondition@" + num2Str(ID) + "/extrapolation_margin"));
    EndnodeReadOnly flightConditionAltitude(EndnodeReadOnly<double>("program_settings/FlightConditions/FlightCondition@" + num2Str(ID) + "/Altitude"));
    EndnodeReadOnly flightConditionMach(EndnodeReadOnly<double>("program_settings/FlightConditions/FlightCondition@" + num2Str(ID) + "/Mach"));
    EndnodeReadOnly flightConditionAllowGridChange(EndnodeReadOnly<bool>("program_settings/FlightConditions/FlightCondition@" + num2Str(ID) + "/allow_grid_change"));
    flightConditionAllowGridChange.read(config);
    flightConditionAltitude.read(config);
    flightConditionMach.read(config);
    flightConditionExtrapolationMargin.read(config);
    flightConditions.push_back(flightCondition());
    flightConditions.back().Mach = flightConditionMach.value();
    flightConditions.back().Altitude = flightConditionAltitude.value();
    flightConditions.back().extrapolationMargin = flightConditionExtrapolationMargin.value();
    flightConditions.back().allowGridChange = flightConditionAllowGridChange.value();
    machNumbers.push_back(flightConditions.back().Mach);
}

void tawCalculatePolarConfig::initializeTrimSettings(const node& config) {
    if (doTrim.value()) {
        if (stabAngleGridMode.value() == "mode_0") {
            stabAngles = {-6.0, 0.0, 6.0};
            neutralStabAngleID = 1;
        } else if (stabAngleGridMode.value() == "mode_1") {
            stabAngles = {-6.0, -4.0, -2.0, 0.0, 2.0, 4.0, 6.0};
            neutralStabAngleID = 3;
        } else if (stabAngleGridMode.value() == "mode_2") {
            std::string iStabString = (EndnodeReadOnly<std::string>("Trim/customStabAngleGrid").read(config)).value();
            std::stringstream ss(iStabString);
            std::string item;
            while (std::getline(ss, item, ';')) {
                stabAngles.push_back(std::stod(item));
            }
        }
    } else {
        stabAngles = {0.0};
        neutralStabAngleID = 0;
    }
    if (doTrim.value() && trimAdditionalCGPositions.value()) {
        CGPositions = {"design", "forward", "aft"};
    } else {
    CGPositions = {"design"};
    }
}

void tawCalculatePolarConfig::read_hl_calibration_factors(const node& config) {
    size_t numberHLPolars = config.getVector(("program_settings/SemiEmpiricalHighLiftAdaptions/hl_polars_correction_values/hl_polar"), 2).size();
    for (size_t HLPolarID(0); HLPolarID < numberHLPolars ; ++HLPolarID) {
        EndnodeReadOnly configurationNode(EndnodeReadOnly<std::string>("SemiEmpiricalHighLiftAdaptions/hl_polars_correction_values/hl_polar@"
                                                            + num2Str(HLPolarID) + "/configuration").read(config));
        EndnodeReadOnly CLMaxNode(EndnodeReadOnly<double>("SemiEmpiricalHighLiftAdaptions/hl_polars_correction_values/hl_polar@"
                                                            + num2Str(HLPolarID) + "/CL_max").read(config));
        EndnodeReadOnly delta_CDNode(EndnodeReadOnly<double>("SemiEmpiricalHighLiftAdaptions/hl_polars_correction_values/hl_polar@"
                                                            + num2Str(HLPolarID) + "/delta_CD").read(config));
        EndnodeReadOnly factor_CDNode(EndnodeReadOnly<double>("SemiEmpiricalHighLiftAdaptions/hl_polars_correction_values/hl_polar@"
                                                            + num2Str(HLPolarID) + "/factor_CD").read(config));
        hl_polar_calibration_values aHLPolarCalibrationContainer;
        aHLPolarCalibrationContainer.CL_max = 0.0;
        aHLPolarCalibrationContainer.delta_CD = 0.0;
        aHLPolarCalibrationContainer.factor_CD = 0.0;
        hl_polar_calibration_data[configurationNode.value()] = aHLPolarCalibrationContainer;
        hl_polar_calibration_data.at(configurationNode.value()).CL_max = CLMaxNode.value();
        hl_polar_calibration_data.at(configurationNode.value()).delta_CD = delta_CDNode.value();
        hl_polar_calibration_data.at(configurationNode.value()).factor_CD = factor_CDNode.value();
    }
}