/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef CALCULATEPOLAR_SRC_TAW_TAWDEFAULTDATA_H_
#define CALCULATEPOLAR_SRC_TAW_TAWDEFAULTDATA_H_

#include <atmosphere/atmosphere.h>
#include <standardFiles/typedefs.h>
#include <aixml/endnode.h>

#include <moduleBasics/runtimeIO.h>

#include <string>
#include <unordered_map>
#include <vector>
#include <map>

/** \brief Class contains all properties of calculated aircraft objects
*/
class tawDefaultData {
 public:
    /* Typedefs */
    /** \brief Constructor of aircraft class
     *  \param &mySettings Reference to calculatePolarSettings object
     */
    explicit tawDefaultData(const std::shared_ptr<RuntimeIO>& rtIO);
    /** \brief Virtual destructor of aircraft class
     */
    virtual ~tawDefaultData() {}

    std::string engineType;         /**< Information about the propulsor type (e.g. fan) [-] */
    double delta_isa;               /**< temperature delta to ISA Standard Atmosphere [K] */
    double initialCruiseMachNumber; /**< Mach number flown in initial cruise phase at initial cruise altitude [-] */
    double initialCruise_Altitude;  /**< Initial cruise altitude [m] */
    /* All EndnodeReadOnly Values are converted automatically to SI-Units/RADIAN when read in */
    double bestCruiseCL;        /**< optimum cruise lift coefficient */
    double designCoGX;          /**< x Position of design center of gravity in [m] */
    double designCoGY;          /**< y Position of design center of gravity in [m] */
    double designCoGZ;          /**< z Position of design center of gravity in [m] */
    double forwardCoGX;         /**< x Position of most forward center of gravity in [m] */
    double forwardCoGY;         /**< y Position of most forward center of gravity in [m] */
    double forwardCoGZ;         /**< z Position of most forward center of gravity in [m] */
    double aftCoGX;             /**< x Position of most afterward center of gravity in [m] */
    double aftCoGY;             /**< y Position of most forward center of gravity in [m] */
    double aftCoGZ;             /**< z Position of most forward center of gravity in [m] */
    double MTOM_CoG_x;          /**< x Position of center of gravity at MTOM in [m] */
    double MTOM_CoG_y;          /**< y Position of center of gravity at MTOM in [m] */
    double MTOM_CoG_z;          /**< z Position of center of gravity at MTOM in [m] */
    double OME_CoG_x;           /**< x Position of center of gravity at OME in [m] */
    double OME_CoG_y;           /**< x Position of center of gravity at OME in [m] */
    double OME_CoG_z;           /**< x Position of center of gravity at OME in [m] */
    double C_L_averageCruise;   /**< average lift coefficient flown during design mission [-] */
    Point designCoG;        /**< Center of gravity coordinates for design conditions at ground */
    Point forwardCoG;       /**< Most forward position of center of gravity coordinates */
    Point aftCoG;           /**< Most afterwards center of gravity coordinates */
    Point MTOM_CoG;         /**< Center of gravity coordinates for MTOM */
    Point OME_CoG;          /**< Center of gravity coordinates for OME */
    Point cruiseTrimRefCoG; /**< Cruise reference center of gravity position */
    size_t numberDevices;
    struct wingDevice {
        std::string type;
        double maxPositiveDeflection;
        double maxNegativeDeflection;
        double innerRelSpanPosition;
        double innerRelFwdChordPosition;
        double innerRelAftChordPosition;
        double outerRelSpanPosition;
        double outerRelFwdChordPosition;
        double outerRelAftChordPosition;
    };
    std::vector<wingDevice> wingDevices;
    Endnode<double>span;
    Endnode<double>MAC;
    Endnode<double>S_ref;
    Endnode<double>neutral_point_xCoordinate;
    Endnode<double>C_L_optimumCruise;
    Endnode<double>C_Dind_optimumCruise;
    Endnode<double>C_Dvisc_optimumCruise;
    Endnode<double>C_Dwave_optimumCruise;
    Endnode<double>C_M_optimumCruise;
    Endnode<double>C_Dtotal_optimumCruise;
    Endnode<double>L_over_D_averageCruise;
    Endnode<double>L_over_D_optimumCruise;
    Endnode<double>CL_max_landing;
    Endnode<double>C_LmaxTO;
    Endnode<double>C_LgroundRoll;
    Endnode<std::string>polar_file;
    Endnode<double>max_spoiler_factor;
    void updateNodes(std::string mode);
 private:
    const std::shared_ptr<RuntimeIO>& rtIO_;                    /**< Pointer to the RuntimeIO */
};

#endif  // CALCULATEPOLAR_SRC_TAW_TAWDEFAULTDATA_H_
